KeystoneRollCall = LibStub('AceAddon-3.0'):NewAddon('KeystoneRollCall', 'AceEvent-3.0', 'AceConsole-3.0','AceTimer-3.0');
_G.KeystoneRollCall = KeystoneRollCall;
KeystoneRollCall.tabs = {}
KeystoneRollCall.AceGUI = LibStub('AceGUI-3.0');	
KeystoneRollCall.Registry = LibStub('AceConfigRegistry-3.0')
KeystoneRollCall.ScrollingTable = LibStub('ScrollingTable');

local _,bag,bagSize,i,msg,dungeonId,key,val
local sortFields = {{"level",false},{"name",true}}
local MSG_PREFIX = "KRC-MSG"
local keystoneEvents = {}
local lastBagUpdate = 0;
local guildSynced, partySynced, raidSynced, instanceSynced = false, false, false, false
KeystoneRollCall.version = GetAddOnMetadata("KeystoneRollCall", "Version");
KeystoneRollCall.newestVersionFound = KeystoneRollCall.version;

KeystoneRollCall.CLASS_COLOR_CODES = {
	["Warrior"] = "|cFFC79C6E",
	["Monk"] = "|cFF558A84",
	["Death Knight"] = "|cFFC41F3B",
	["Shaman"] = "|cFF0070DE",
	["Priest"] = "|cFFFFFFFF",
	["Rogue"] = "|cFFFFF569",
	["Hunter"] = "|cFFABD473",
	["Paladin"] = "|cFFF58CBA",
	["Warlock"] = "|cFF9482C9",
	["Mage"] = "|cFF69CCF0",
	["Druid"] = "|cFFFF7D0A",
	["Demon Hunter"] = "|cFFA330C9"
}

KeystoneRollCall.columnSort = function (table, rowa, rowb, sortbycol)
	local field = table.cols[sortbycol].field or table.cols[sortbycol].name:lower();
	local asc = table.cols[sortbycol].sort == "asc"
	local ka = KeystoneRollCall.tableData[rowa][1]
	local kb = KeystoneRollCall.tableData[rowb][1]
	return KeystoneRollCall:CompareKeystones(field, asc, ka, kb);
end

function KeystoneRollCall:getGroupChatChannel()
	if IsInGroup(LE_PARTY_CATEGORY_INSTANCE) then
		return "INSTANCE_CHAT";
	elseif UnitInRaid("player") then
		return "RAID";
	elseif IsInGroup() then
		return "PARTY";
	else
		return "SAY";
	end
end

function KeystoneRollCall:getDungeonName(dungeon)
	return dungeon and C_ChallengeMode.GetMapInfo(dungeon) or ""
end

function KeystoneRollCall:keystoneToString(keystone)
	if keystone.level == 0 then
		return "No keystone found for " .. keystone.name
	end
	return "+" .. keystone.level .. " "  ..  KeystoneRollCall:getDungeonName(keystone.dungeon) .. "  -  " ..  keystone.name
end


function KeystoneRollCall:SendKeystone(requestSync,keystone,channel,target)
	if keystone then
		local msg = table.concat({requestSync, self:extractKeystone(keystone)}, ",") 
		SendAddonMessage(MSG_PREFIX, msg, channel, target)
	end
end

function KeystoneRollCall:ReportAvailableKeys(channel,target)
	self:DoSort()
	SendChatMessage("Keystone Roll-Call (Personal Keys):",channel, "Common",target)
	for i, keystone in pairs(self.db.global.keystones) do
		if keystone.isPlayers and keystone.dungeon and tonumber(keystone.dungeon) > 0 then
			SendChatMessage("   " .. self:keystoneToString(keystone), channel, "Common", target)
		end
	end
end

function KeystoneRollCall:ReportPartyKey(channel,target,unit)

	if UnitExists(unit) and UnitIsPlayer(unit) and GetUnitName(unit, true) and UnitClass(unit) then
		local keystone = KeystoneRollCall:getKeystoneByUnit(unit)
		if keystone then
			SendChatMessage("   " .. self:keystoneToString(keystone), channel, "Common", target)
		end
	end
end

function KeystoneRollCall:ReportPartyKeys(channel,target)
	self:DoSort()
	SendChatMessage("Keystone Roll-Call (Party/Raid Keys):",channel, "Common",target)
	if UnitInRaid("player") then
		for i = 1,40 do
			self:ReportPartyKey(channel,target,"raid"..i)
		end
	elseif IsInGroup() then
		for i = 1,5 do
			self:ReportPartyKey(channel,target,"party"..i)
		end
	else
		self:ReportPartyKey(channel,target,"player")
	end
end

function KeystoneRollCall:LocalReportAvailableKeys()
	print("Keystone Roll-Call (" .. #self.db.global.keystones .. ")")
	for i, keystone in ipairs(self.db.global.keystones) do
		if keystone.dungeon and tonumber(keystone.dungeon) > 0 and keystone.isPlayers then
			print(KeystoneRollCall:colorLevelDifficulty(keystone.level,keystone.level).. " |cFFFFFF00" .. KeystoneRollCall:getDungeonName(keystone.dungeon) .. "|r    " ..(keystone.name or ""))
		end
	end
end

function KeystoneRollCall:OnInitialize() 
	self:RegisterChatCommand("krc", "SlashCommand")
	self:RegisterSettings()
	
end


function KeystoneRollCall:SlashCommand(args)
	if args == "version" then
		print(KeystoneRollCall.version);
	else
		if not KeystoneRollCall.frame or not KeystoneRollCall.frame:IsVisible() then
			KeystoneRollCall:Show()
		end
	end
end

function KeystoneRollCall:RequestKeystoneSync()		
	local keystone = self:getPlayerKeystone()
	local groupType = self:getGroupChatChannel()

	if groupType == "INSTANCE_CHAT" then
		self:SendKeystone(instanceSynced and "0" or "1",keystone,"INSTANCE_CHAT")
		--print("Requesting Instance Sync",instanceSynced and "0" or "1")
		instanceSynced = true
	elseif groupType == "RAID" then
		self:SendKeystone(raidSynced and "0" or "1",keystone,"RAID")
		--print("Requesting Instance Sync",instanceSynced and "0" or "1")
		instanceSynced = true
	elseif groupType == "PARTY" then
		self:SendKeystone(partySynced and "0" or "1",keystone,"PARTY")
		--print("Requesting Instance Sync",instanceSynced and "0" or "1")
		instanceSynced = true
	end
	
	if  GetGuildInfo("player")  then
		self:SendKeystone(guildSynced and "0" or "1",keystone,"GUILD")
		--print("Requesting Guild Sync",guildSynced and "0" or "1")
		guildSynced = true
	end
	 
end

function KeystoneRollCall:AnnounceNewPlayerKeystone(keystone,oldKeystone)
	if keystone then
		local isNew = keystone.level and keystone.level > 0 and (not oldKeystone or keystone.level ~= oldKeystone.level or keystone.dungeon ~= oldKeystone.dungeon)
		if isNew then
			if KeystoneRollCall.db.global.printPlayerNewKey then
				print("You have a new keystone! +" ..  KeystoneRollCall:colorLevelDifficulty(keystone.level,keystone.level) .. " "..KeystoneRollCall:getDungeonName(keystone.dungeon) )
			end
			if KeystoneRollCall.db.global.partyPlayerNewKey and IsInGroup() then
				SendChatMessage(UnitName("Player") .. " has a new keystone! +" .. keystone.level .. " " .. KeystoneRollCall:getDungeonName(keystone.dungeon), "PARTY", "Common")
			end
			if KeystoneRollCall.db.global.guildPlayerNewKey and GetGuildInfo("player") then
				SendChatMessage(UnitName("Player") .. " has a new keystone! +" .. keystone.level .. " " .. KeystoneRollCall:getDungeonName(keystone.dungeon), "GUILD", "Common")
			end
		end
	end
end

function KeystoneRollCall:AnnounceNewPlayerBestRun(keystone,oldKeystone)
	if keystone then
		local isNewBest = keystone.bestLevel and keystone.bestLevel > 0 and (not oldKeystone or keystone.bestLevel > oldKeystone.bestLevel)
		if isNewBest then
			if KeystoneRollCall.db.global.printPlayerBestRun then
				print("You have achieved a new best mythic+ run for the week! +" .. keystone.bestLevel .. " "..KeystoneRollCall:getDungeonName(keystone.bestDungeon).."" )
			end
			if KeystoneRollCall.db.global.partyPlayerBestRun and IsInGroup() then
				SendChatMessage(UnitName("Player") .. " has achieved a new best mythic+ run for the week! +" .. keystone.bestLevel .. " " .. KeystoneRollCall:getDungeonName(keystone.bestDungeon), "PARTY", "Common")
			end
			if KeystoneRollCall.db.global.guildPlayerBestRun and GetGuildInfo("player") then
				SendChatMessage(UnitName("Player") .. " has achieved a new best mythic+ run for the week! +" .. keystone.bestLevel .. " " .. KeystoneRollCall:getDungeonName(keystone.bestDungeon), "GUILD", "Common")
			end
		end
	end
end

local function processAddonMessage(prefix,msg,channel,sender) 
	if prefix == MSG_PREFIX then
	
		if sender == KeystoneRollCall.playerName then
			return
		end
		
		
		local msgParts =  {strsplit(",",msg)};
		if msgParts[1] == "1" then
			for i, keystone in ipairs(KeystoneRollCall.db.global.keystones) do
				if keystone.isPlayers then
					KeystoneRollCall:SendKeystone(0,keystone,channel,sender)
				end
			end
		end
	
		tremove(msgParts,1)
		
		local keystone, oldKeystone = KeystoneRollCall:buildKeystone(msgParts,sender)

		if keystone then		
			if KeystoneRollCall.db.global.printOthersBestRuns and keystone.bestLevel and keystone.bestLevel > 0 and (not oldKeystone or keystone.bestLevel > oldKeystone.bestLevel) then
				print(KeystoneRollCall:colorClass(keystone.name,keystone.class) .. " has achieved a new best mythic+ run for the week! +" .. KeystoneRollCall:colorLevelDifficulty(keystone.bestLevel) .. " "..KeystoneRollCall:getDungeonName(keystone.bestDungeon).."" )
			end
		
			if KeystoneRollCall.db.global.printOthersNewKeys and keystone.level and keystone.level > 0 and (not oldKeystone or keystone.level ~= oldKeystone.level or keystone.dungeon ~= oldKeystone.dungeon) then
				print(KeystoneRollCall:colorClass(keystone.name,keystone.class) .. " has a new keystone! +" ..  KeystoneRollCall:colorLevelDifficulty(keystone.level,keystone.level) .. " "..KeystoneRollCall:getDungeonName(keystone.dungeon) )
			end
			
			if keystone.version and tonumber(keystone.version) and tonumber(keystone.version) > tonumber(KeystoneRollCall.newestVersionFound) then
				KeystoneRollCall.newestVersionFound = keystone.version
				print("A newer version (|cFF69CCF0".. keystone.version .. "|r) of |cFFFFFF00Keystone Roll-Call|r has been detected, please update the addon to ensure you can continue to send and receive keystone data with others.")
			end
		end
	end
end


function testNewKeys()
	local keystone = {}
	keystone.name = "NOONE"..time()
	keystone.class = "Druid"
	keystone.faction = "0"
	keystone.dungeon = 200
	keystone.level = 12
	keystone.affix1 = 0
	keystone.affix2 = 0
	keystone.affix3 = 0
	keystone.bestLevel = 14
	keystone.bestDungeon = 200
	keystone.isPlayers = false;
	
	
	local msg = table.concat({"0", KeystoneRollCall:extractKeystone(keystone)}, ",") ..",1.3"
	processAddonMessage(MSG_PREFIX,msg,"SAY","NOONE")
end


local function EventFunc(self, event, ...)
	--print(guildSynced, partySynced, raidSynced, instanceSynced)
	if event ~= "CHAT_MSG_ADDON" and event ~= "BAG_UPDATE" then
		--print(event, select(1,...))
	end
	if event == "PLAYER_LEAVING_WORLD" then
		KeystoneRollCall.loaded = false;
	elseif event == "PLAYER_LOGIN" then
		guildSynced, partySynced, raidSynced, instanceSynced = false, false, false, false
		local name, realm = UnitFullName("player")
		KeystoneRollCall.playerName = name .."-" .. realm
		KeystoneRollCall.playerRealm = realm
		RegisterAddonMessagePrefix(MSG_PREFIX)
	elseif event == "GROUP_JOINED" then
		partySynced, raidSynced, instanceSynced = false, false, false
		C_ChallengeMode.RequestMapInfo();
		C_ChallengeMode.RequestRewards();
	elseif event == "PLAYER_ENTERING_WORLD" or (event == "BAG_UPDATE" and KeystoneRollCall.loaded) then
		if event ~= "BAG_UPDATE" or lastBagUpdate ~= time() then
			lastBagUpdate = time() 
			C_ChallengeMode.RequestMapInfo();
			C_ChallengeMode.RequestRewards();
		end
	elseif event == "CHALLENGE_MODE_MAPS_UPDATE" or event == "CHALLENGE_MODE_NEW_RECORD" then
		KeystoneRollCall.loaded = true;
		local keystone,oldKeystone = KeystoneRollCall:buildPlayerKeystone()
		if keystone then
			guildSynced, partySynced, raidSynced, instanceSynced = false, false, false, false
			KeystoneRollCall:AnnounceNewPlayerBestRun(keystone,oldKeystone)
			KeystoneRollCall:AnnounceNewPlayerKeystone(keystone,oldKeystone)
		end
		
		KeystoneRollCall:RequestKeystoneSync()
	elseif event == "CHAT_MSG_ADDON" then
		processAddonMessage(...)
	end
end

local eventHandler = CreateFrame("Frame", nil, WorldFrame)
eventHandler:SetScript("OnEvent", EventFunc)
eventHandler:RegisterEvent("PLAYER_LEAVING_WORLD")
eventHandler:RegisterEvent("PLAYER_LOGIN")
eventHandler:RegisterEvent("PLAYER_ENTERING_WORLD")
eventHandler:RegisterEvent("CHALLENGE_MODE_NEW_RECORD")
eventHandler:RegisterEvent("GROUP_JOINED")
eventHandler:RegisterEvent("CHALLENGE_MODE_MAPS_UPDATE")
eventHandler:RegisterEvent("BAG_UPDATE")
eventHandler:RegisterEvent("CHAT_MSG_ADDON")




function KeystoneRollCall:CompareKeystones(field,asc,ka,kb)
	if sortFields[1][1] == field then
		sortFields[1][2] = asc
	else
		for i = 2, #sortFields do
			if sortFields[i] and sortFields[i][1] == field then
				tremove(sortFields,i)
			end
		end
		tinsert(sortFields,1,{field,true})
	end

	for i = 1, #sortFields do
		
		local av
		local bv
		if sortFields[i][1] == "dungeon" then
			av = KeystoneRollCall:getDungeonName(ka.dungeon) or "ZZZ"
			bv = KeystoneRollCall:getDungeonName(kb.dungeon) or "ZZZ"
		elseif sortFields[i][1] == "realm" then
			local nameParts =  {strsplit("-",ka.name or"")}
			av = nameParts[2] or""
			nameParts =  {strsplit("-",kb.name or"")}
			bv = nameParts[2] or""
		elseif sortFields[i][1] == "bestLevel" or sortFields[i][1] == "level" or sortFields[i][1] == "ilvl" or sortFields[i][1] == "equippedIlvl" then
			av =  ka[sortFields[i][1]] or 0
			bv =  kb[sortFields[i][1]] or 0
		else
			av =  ka[sortFields[i][1]] or ""
			bv =  kb[sortFields[i][1]] or ""
		end
		if av < bv then
			return sortFields[i][2];
		elseif av > bv then
			return not sortFields[i][2];
		end
	end
	return 
end

function KeystoneRollCall:DoSort()
	table.sort(self.db.global.keystones, function(ka,kb) 
		return self:CompareKeystones(sortFields[1][1],sortFields[1][2],ka,kb)
	end);
end


function KeystoneRollCall:Hide()
	if self.frame and self.frame:IsVisible() then
		self.frame:Hide()
	end
end

function KeystoneRollCall:Show()
	if self.frame and self.frame:IsVisible() then
		return
	end

	GuildRoster()
	
	local frame = KeystoneRollCall.AceGUI:Create("Frame")
	_G["KeystoneRollCall_Display"] = frame
	tinsert(UISpecialFrames, "KeystoneRollCall_Display") 
	KeystoneRollCall.frame = frame
	frame:SetTitle("Keystone Roll-Call")
	frame:SetStatusText("")
	frame:SetWidth(760)
	frame:EnableResize(false);
	frame:SetCallback("OnClose", function(widget) KeystoneRollCall.AceGUI:Release(widget) end)
	frame:SetLayout("Fill")
	
	local tabDef = KeystoneRollCall.tabs["personal"]
	local tabList = {{text = tabDef.text, value = tabDef.value, index = tabDef.index}}

	if self.db.global.partyTabEnabled then
		local tabDef = KeystoneRollCall.tabs["party"]
		tinsert(tabList,{text = tabDef.text, value = tabDef.value, index = tabDef.index})
	end
	if self.db.global.guildOnlineTabEnabled then
		local tabDef = KeystoneRollCall.tabs["guild_online"]
		tinsert(tabList,{text = tabDef.text, value = tabDef.value, index = tabDef.index})
	end
	if self.db.global.guildAllTabEnabled then
		local tabDef = KeystoneRollCall.tabs["guild_all"]
		tinsert(tabList,{text = tabDef.text, value = tabDef.value, index = tabDef.index})
	end
	
	table.sort(tabList,function(a,b)
		return a.index < b.index
	end)

	local tab =  KeystoneRollCall.AceGUI:Create("TabGroup")
	tab:SetLayout("Flow")
	tab:SetTabs(tabList)
	tab:SetCallback("OnGroupSelected", function(container, event, group)
		container:ReleaseChildren()
		local tab = KeystoneRollCall.tabs[group]
		tab.setup(container);
		KeystoneRollCall:DoSort();
		tab.populate();
	end)
	tab:SelectTab("personal")
	frame:AddChild(tab)
end